Puppet::Type.type(:logical_volume).provide :lvm do
  desc 'Manages LVM logical volumes on Linux'

  confine kernel: :linux

  commands lvcreate: 'lvcreate',
           lvremove: 'lvremove',
           lvresize: 'lvresize',
           lvextend: 'lvextend',
           lvs: 'lvs',
           vgs: 'vgs',
           mkswap: 'mkswap',
           swapoff: 'swapoff',
           swapon: 'swapon',
           umount: 'umount',
           blkid: 'blkid',
           dmsetup: 'dmsetup',
           lvconvert: 'lvconvert',
           lvdisplay: 'lvdisplay',
           lsblk: 'lsblk',
           dd: 'dd',
           fsadm: 'fsadm'

  def round_to_extent(size)
    lvm_size_units = {
      "K" => 1**0, "M" => 1024**1, "G" => 1024**2, "T" => 1024**3, "P" => 1024**4, "E" => 1025**5,
    }

    if @resource[:size] =~ /^([0-9]+(\.[0-9]+)?)([KMGTPE])/i
      size_value = $1.to_f
      size_unit  = $3.upcase
      size_kibi = (size_value * lvm_size_units[size_unit]).to_i
      if vgs('--noheading', '-o', 'vg_extent_size', '--units', 'k', "#{@resource[:volume_group]}") =~ /\s+(\d+)\.\d+k/i
        vg_extent_size_kibi = $1.to_i
      end
      new_size_kibi = ((size_kibi + vg_extent_size_kibi - 1) / vg_extent_size_kibi) * vg_extent_size_kibi
      "#{new_size_kibi}k"
    else
      size
    end
  end

  def self.instances
    get_logical_volumes.map do |logical_volumes_line|
      logical_volumes_properties = get_logical_volume_properties(logical_volumes_line)
      instance = new(logical_volumes_properties)
      # Save the volume group in the provider so the type can find it
      instance.volume_group = logical_volumes_properties[:volume_group]
      instance
    end
  end

  def self.get_logical_volumes
    full_lvs_output = lvs.split("\n")

    # Remove first line
    logical_volumes = full_lvs_output.drop(1)

    logical_volumes
  end

  def self.get_logical_volume_properties(logical_volumes_line)
    logical_volumes_properties = {}

    # lvs output formats thus:
    # LV      VG       Attr       LSize   Pool Origin Data%  Meta%  Move Log Cpy%Sync Convert

    # Split on spaces
    output_array = logical_volumes_line.gsub(%r{\s+}m, ' ').strip.split(' ')

    # Assign properties based on headers
    logical_volumes_properties[:ensure]       = :present
    logical_volumes_properties[:name]         = output_array[0]
    logical_volumes_properties[:volume_group] = output_array[1]

    logical_volumes_properties
  end

  # Just assume that the volume group is correct, we don't support changing
  # it anyway.
  attr_writer :volume_group

  def volume_group
    @resource ? @resource[:volume_group] : @volume_group
  end

  def create
    args = ['-y', '-Wy', '-Zy']

    args.push('-n', @resource[:name]) unless @resource[:thinpool] == true

    size_option = '--size'
    if @resource[:thinpool].is_a? String
      size_option = '--virtualsize'
    end

    if @resource[:size]
      size = @resource[:size]
      if size == 'max'
          size = vgs('--noheading', '-o', 'vg_size', '--units', 'k', "#{@resource[:volume_group]}").strip
      elsif @resource[:round_to_extent] then
          size = round_to_extent(size)
      end
      args.push(size_option, size)
    elsif @resource[:initial_size]
      args.push('--size',
        if @resource[:round_to_extent] then
            round_to_extent(@resource[:initial_size])
        else
            @resource[:initial_size]
        end)
    end
    if @resource[:extents]
      args.push('--extents', @resource[:extents])
    end

    if !@resource[:extents] && !@resource[:size] && !@resource[:initial_size]
      args.push('--extents', '100%FREE')
    end

    if @resource[:stripes]
      args.push('--stripes', @resource[:stripes])
    end

    if @resource[:stripesize]
      args.push('--stripesize', @resource[:stripesize])
    end

    if @resource[:poolmetadatasize]
      args.push('--poolmetadatasize', @resource[:poolmetadatasize])
    end

    if @resource[:mirror]
      args.push('--mirrors', @resource[:mirror])
      if @resource[:mirrorlog]
        args.push('--mirrorlog', @resource[:mirrorlog])
      end
      if @resource[:region_size]
        args.push('--regionsize', @resource[:region_size])
      end
      if @resource[:no_sync]
        args.push('--nosync')
      end
    end

    if @resource[:alloc]
      args.push('--alloc', @resource[:alloc])
    end

    if @resource[:readahead]
      args.push('--readahead', @resource[:readahead])
    end

    if @resource[:persistent]
      # if persistent param is true, set arg to "y", otherwise set to "n"
      args.push('--persistent', [:true, true, 'true'].include?(@resource[:persistent]) ? 'y' : 'n')
    end

    if @resource[:minor]
      args.push('--minor', @resource[:minor])
    end

    if @resource[:type]
      args.push('--type', @resource[:type])
    end

    if @resource[:thinpool]
      args.push('--thin')
      args << if @resource[:thinpool].is_a? String
                @resource[:volume_group] + '/' + @resource[:thinpool]
              else
                @resource[:volume_group] + '/' + @resource[:name]
              end
    else
      args << @resource[:volume_group]
    end
    info("Calling lvcreate with args: #{args}")
    lvcreate(*args)
    lvzero
  end

  def destroy
    name_escaped = "#{@resource[:volume_group].gsub('-', '--')}-#{@resource[:name].gsub('-', '--')}"
    if blkid(path) =~ %r{\bTYPE=\"(swap)\"}
      swapoff(path)
    end
    dmsetup('remove', name_escaped)
    lvremove('-f', path)
  end

  def exists?
    lvs(@resource[:volume_group]) =~ lvs_pattern
  rescue Puppet::ExecutionFailure
    # lvs fails if we give it an empty volume group name, as would
    # happen if we were running `puppet resource`. This should be
    # interpreted as "The resource does not exist"
    nil
  end

  def exec_cmd(*cmd)
    output = Puppet::Util::Execution.execute(cmd, :failonfail => false, :combine => true)
    {:out => output, :exit => $CHILD_STATUS.exitstatus}
  end

  def size
    unit = if @resource[:size] =~ %r{^\d+\.?\d{0,2}([KMGTPE])}i
             Regexp.last_match(1).downcase
           else
             # If we are getting the size initially we don't know what the
             # units will be, default to KB
             'k'
           end

    raw = lvs('--noheading', '--unit', unit, path)

    if raw =~ %r{\s+(\d+)\.(\d+)#{unit}}i
      if Regexp.last_match(2).to_i.zero?
        return Regexp.last_match(1) + unit.capitalize
      else
        return Regexp.last_match(1) + '.' + Regexp.last_match(2).sub(%r{0+$}, '') + unit.capitalize
      end
    end
  end

  def size=(new_size)
    lvm_size_units = { 'K' => 1, 'M' => 1024, 'G' => 1024**2, 'T' => 1024**3, 'P' => 1024**4, 'E' => 1024**5 }

    current_size = size

    if current_size =~ %r{^([0-9]+(\.[0-9]+)?)([KMGTPE])}i
      current_size_value = Regexp.last_match(1).to_f
      current_size_unit = Regexp.last_match(3).upcase
      current_size = (current_size_value * lvm_size_units[current_size_unit]).to_i
    end

    info( "Current: value=#{current_size_value}, unit=#{current_size_unit}, kibi=#{current_size}" )

    if new_size == 'max'
      new_size = vgs('--noheading', '-o', 'vg_size', '--units', 'k', "#{@resource[:volume_group]}").strip
    end

    if new_size =~ %r{^([0-9]+(\.[0-9]+)?)([KMGTPE])}i
      new_size_value = Regexp.last_match(1).to_f
      new_size_unit = Regexp.last_match(3).upcase
      new_size = (new_size_value * lvm_size_units[new_size_unit]).to_i
    end

    info( "New: value=#{new_size_value}, unit=#{new_size_unit}, kibi=#{new_size}" )

    ## Get the extend size
    if lvs('--noheading', '-o', 'vg_extent_size', '--units', 'k', path) =~ %r{\s+(\d+)\.\d+k}i
      vg_extent_size = Regexp.last_match(1).to_i
    end

    if new_size < current_size
      if @resource[:size_is_minsize] == :true or @resource[:size_is_minsize] == true or @resource[:size_is_minsize] == 'true'
        info( "Logical volume already has minimum size of #{new_size} (currently #{current_size})" )
      else
        if not @resource[:allow_reduce]
          fail( "Decreasing the size requires manual intervention (#{new_size} < #{current_size})" )
        end
        if new_size % vg_extent_size != 0
          if @resource[:round_to_extent]
            new_size = ((new_size + vg_extent_size - 1) / vg_extent_size) * vg_extent_size
            if new_size >= current_size
              info( "Logical volume already has a size of #{current_size}" )
              return
            end
          else
            fail( "Cannot reduce to size #{new_size} because VG extent size is #{vg_extent_size} KB" )
          end
        end
        exec_cmd('umount', path)
        exec_cmd('fsadm', '-y', 'check', path )
        r = exec_cmd('fsadm', '-y', 'resize', path, "#{new_size}k")
        if r[:exit] != 0 and [:true, "true", true ].include? @resource[:nuke_fs_on_resize_failure]
          info( "Failed 'fsadm resize' erase the disk #{r}" )
          exec_cmd('dd', 'if=/dev/zero', "of=#{path}", "bs=512", "count=16", "conv=notrunc")
          blkid('-g')
        end
        r = exec_cmd('lvresize', '-r', '-f', '-L', "#{new_size}k", path)
        if r[:exit] != 0
          if [:true, "true", true ].include? @resource[:nuke_fs_on_resize_failure]
            info( "Failed 'fsadm resize' erase the disk #{r}" )
            exec_cmd('dd', 'if=/dev/zero', "of=#{path}", "bs=512", "count=16", "conv=notrunc")
            blkid('-g')
            lvresize( '-f', '-L', "#{new_size}k", path) || fail( "Cannot reduce to size #{new_size} because lvresize failed." )
          else
            fail( "Cannot reduce to size #{new_size} because lvresize failed." )
          end
        end
      end
    elsif new_size > current_size
      if new_size % vg_extent_size != 0
        if @resource[:round_to_extent]
          new_size = ((new_size + vg_extent_size - 1) / vg_extent_size) * vg_extent_size
          if new_size <= current_size
            info( "Logical volume already has a size of #{current_size}" )
            return
          end
        else
          fail( "Cannot extend to size #{new_size} because VG extent size is #{vg_extent_size} KB" )
        end
      end
      lvextend( '-L', "#{new_size}k", path) || fail( "Cannot extend to size #{new_size} because lvextend failed." )
      exec_cmd("seek_end=$(($(blockdev --getsz #{path})/2048 - 10)); dd if=/dev/zero of=#{path} bs=1M seek=${seek_end} count=10")
      exec_cmd('umount', path)
      exec_cmd('fsadm', '-y', 'check', path )
      r = exec_cmd('fsadm', '-y', 'resize', path, "#{new_size}k")
      if r[:exit] != 0 and [:true, "true", true ].include? @resource[:nuke_fs_on_resize_failure]
        info( "Failed 'fsadm resize' erase the disk #{r}" )
        exec_cmd('dd', 'if=/dev/zero', "of=#{path}", "bs=512", "count=16", "conv=notrunc")
        blkid('-g')
      end
    else
      info( "Logical volume already has a size of #{current_size}" )
    end
  end

  # Look up the current number of mirrors (0=no mirroring, 1=1 spare, 2=2 spares....). Return the number as string.
  def mirror
    raw = lvdisplay(path)
    # If the first attribute bit is "m" or "M" then the LV is mirrored.
    if raw =~ %r{Mirrored volumes\s+(\d+)}im
      # Minus one because it says "2" when there is only one spare. And so on.
      n = Regexp.last_match(1).to_i - 1
      # puts " current mirrors: #{n}"
      return n.to_s
    end
    0.to_s
  end

  def mirror=(new_mirror_count)
    current_mirrors = mirror.to_i
    if new_mirror_count.to_i != current_mirrors
      puts "Change mirror from #{current_mirrors} to #{new_mirror_count}..."
      args = ['-m', new_mirror_count]
      if @resource[:mirrorlog]
        args.push('--mirrorlog', @resource[:mirrorlog])
      end

      # Region size cannot be changed on an existing mirror (not even when changing to zero mirrors).

      if @resource[:alloc]
        args.push('--alloc', @resource[:alloc])
      end
      args.push(path)
      lvconvert(*args)
    end
  end

  # Location of the mirror log. Empty string if mirror==0, else "mirrored", "disk" or "core".
  def mirrorlog
    vgname = (@resource[:volume_group]).to_s
    lvname = (@resource[:name]).to_s
    raw = lvs('-a', '-o', '+devices', vgpath)

    if mirror.to_i > 0
      if raw =~ %r{\[#{lvname}_mlog\]\s+#{vgname}\s+}im
        if raw =~ %r{\[#{lvname}_mlog\]\s+#{vgname}\s+mw\S+}im # attributes start with "m" or "M"
          return 'mirrored'
        else
          return 'disk'
        end
      else
        return 'core'
      end
    end
    nil
  end

  def mirrorlog=(new_mirror_log_location)
    # It makes no sense to change this unless we use mirrors.
    mirror_count = mirror.to_i
    if mirror_count.to_i > 0
      current_log_location = mirrorlog.to_s
      if new_mirror_log_location.to_s != current_log_location
        # puts "Change mirror log location to #{new_mirror_log_location}..."
        args = ['--mirrorlog', new_mirror_log_location]
        if @resource[:alloc]
          args.push('--alloc', @resource[:alloc])
        end
        args.push(path)
        lvconvert(*args)
      end
    end
  end

    private

  def lvs_pattern
    # lvs output format:
    # LV      VG       Attr       LSize   Pool Origin Data%  Meta%  Move Log Cpy%Sync Convert
    %r{\s+#{Regexp.quote @resource[:name]}\s+#{Regexp.quote @resource[:volume_group]}\s+}
  end

  def path
    "/dev/#{@resource[:volume_group]}/#{@resource[:name]}"
  end

  # Device path of only the volume group (does not include the logical volume).
  def vgpath
    "/dev/#{@resource[:volume_group]}"
  end

  def lvzero
    if lvs('--noheading', '-o', 'lv_size', '--units', 'm', path) =~ /\s+(\d+)\.\d+m/i
      lv_size = $1.to_i
      lv_size = lv_size - 2
      begin
        dd('if=/dev/zero', 'of=' + path, 'bs=1M', "seek=#{lv_size}")
      rescue
      end
      begin
        dd('if=/dev/zero', 'of=' + path, 'bs=1M', 'count=100')
      rescue
      end
    end
  end
end
