require 'csv'

Puppet::Type.type(:volume_group).provide :lvm do
  desc 'Manages LVM volume groups on Linux'

  confine kernel: :linux

  commands vgcreate: 'vgcreate',
           vgremove: 'vgremove',
           pvremove: 'pvremove',
           vgs: 'vgs',
           vgextend: 'vgextend',
           vgreduce: 'vgreduce',
           vgscan: 'vgscan',
           pvs: 'pvs',
           lvremove: 'lvremove',
           umount: 'umount'

  def self.instances
    get_volume_groups.map do |volume_groups_line|
      volume_groups_properties = get_logical_volume_properties(volume_groups_line)
      new(volume_groups_properties)
    end
  end

  def self.get_volume_groups
    full_vgs_output = vgs.split("\n")

    # Remove first line
    volume_groups = full_vgs_output.drop(1)

    volume_groups
  end

  def self.get_logical_volume_properties(volume_groups_line)
    volume_groups_properties = {}

    # vgs output formats thus:
    # VG       #PV #LV #SN Attr   VSize  VFree

    # Split on spaces
    output_array = volume_groups_line.gsub(%r{\s+}m, ' ').strip.split(' ')

    # Assign properties based on headers
    # Just doing name for now...
    volume_groups_properties[:ensure]     = :present
    volume_groups_properties[:name]       = output_array[0]

    volume_groups_properties
  end

  def create
    vgcreate(@resource[:name], *@resource.should(:physical_volumes))
  end

  def destroy
    vgremove(@resource[:name])
  end

  def exists?
    vgs(@resource[:name])
  rescue Puppet::ExecutionFailure
    false
  end

  def exec_cmd(*cmd)
    output = Puppet::Util::Execution.execute(cmd, :failonfail => false, :combine => true)
    {:out => output, :exit => $CHILD_STATUS.exitstatus}
  end

  def physical_volumes=(new_volumes = [])
    # Only take action if createonly is false just to be safe
    #  this is really only here to enforce the createonly setting
    #  if something goes wrong in physical_volumes
    if @resource[:createonly].to_s == 'false'
      existing_volumes = physical_volumes
      extraneous = existing_volumes - new_volumes
      pv_to_lv={}
      pv_to_dev={}
      csv = CSV.new(pvs('-o', 'pv_name,vg_name,lv_name', '--separator', ','),
                   :headers => true, :header_converters => :symbol)
      csv.to_a.map {|row| row.to_hash}.each do |m|
        unless m[:lv].nil?
          pv_to_lv[m[:pv].strip] = "#{m[:vg]}/#{m[:lv]}"
          pv_to_dev[m[:pv].strip] = "#{m[:vg].gsub('-','--')}-#{m[:lv].gsub('-','--')}"
        end
      end

      if extraneous == existing_volumes
        extraneous.each do |volume|
          if pv_to_lv.has_key?(volume)
            exec_cmd('/bin/umount', "/dev/mapper/#{pv_to_dev[volume]}")
            lvremove('-f', pv_to_lv[volume])
          end
        end
        vgremove(@resource[:name], '--force')
        extraneous.each do |volume|
            pvremove(volume)
        end
        vgcreate(@resource[:name], *new_volumes)
      else
        extraneous.each do |volume|
          if pv_to_lv.has_key?(volume)
            exec_cmd('/bin/umount', "/dev/mapper/#{pv_to_dev[volume]}")
            lvremove('-f', pv_to_lv[volume])
          end
          reduce_with(volume)
          pvremove(volume)
        end
        missing = new_volumes - existing_volumes
        missing.each { |volume| extend_with(volume) }
      end
    end
  end

  def physical_volumes
    if @resource[:createonly].to_s == 'false' || !vgs(@resource[:name])
      lines = `pvs -o pv_name,vg_name --separator ',' | awk -F ',' 'NR>1{cmd="find -L /dev/disk/by-{path,id/wwn-*} -samefile" $1; cmd | getline $1;print $1 "," $2; next};{print}' 2>/dev/null`
      lines.split(%r{\n}).grep(%r{,#{@resource[:name]}$}).map do |s|
        s.split(%r{,})[0].strip
      end
    else
      # Trick the check by setting the returned value to what is
      #  listed in the puppet catalog
      @resource[:physical_volumes]
    end
  end

    private

  def reduce_with(volume)
    vgreduce(@resource[:name], volume)
  rescue Puppet::ExecutionFailure => detail
    raise Puppet::Error, "Could not remove physical volume #{volume} from volume group '#{@resource[:name]}'; this physical volume may " \
        + "be in use and may require a manual data migration (using pvmove) before it can be removed (#{detail.message})"
  end

  def extend_with(volume)
    vgextend(@resource[:name], volume)
  rescue Puppet::ExecutionFailure => detail
    raise Puppet::Error, "Could not extend volume group '#{@resource[:name]}' with physical volume #{volume} (#{detail.message})"
  end
end
